{ (C) 2001 Marek Wodzinski
 This code is licensed under GNU GPL
 version:0.2 alpha
 Last modified: Tue Sep 25 17:54:00 2001

 0.1a -> 0.2a : added find_db_ids and destroy_db_ids to find all
                databases in REX
}

unit rex;
interface

type
  //buffers
  tbufor=array[0..32767] of byte;
  twbufor=^tbufor;

  //database fields types
  twdbtype=^tdbtype;
  tdbtype=record
            typ:byte;
	    size:byte;
	    next:twdbtype;
	  end;

  //database records
  twdbrecord=^tdbrecord;
  tdbrecord=record
              size:integer;
	      data:twbufor;
	      next:twdbrecord;
	    end;

  //databases
  twdbnr=^tdbnr;
  tdbnr=record
          nr:cardinal;
	  next:twdbnr;
	end;

//low level (serial port communication)
function open_port(nazwa:string):boolean;
procedure close_port;
procedure wyslij_komende(s:string);
procedure wyslij_do_portu(buf:pointer;chars:word);
function czytaj_wynik(buf:twbufor;max_chars:word;timeout:word):word;

//rex level (commands to comunicate with rex)
function open_rex(nazwa:string):boolean;
procedure close_rex;
function dssr(buf:twbufor;page:word):integer;
function read_page(buf:twbufor;page:word):longint;
procedure dssw(buf:twbufor;page:word;chars:integer);
procedure dssg(buf:twbufor);
procedure dssf(buf:twbufor);
function dstr(buf:twbufor):string;
procedure dstw(rextime:string);

//helpers (print buffer, make checksums etc.)
function make_cksum(k:string):string;
function make_cksum_bin(buf:twbufor;chars:integer):string;
function buftostr(buf:twbufor;chars:word):string;
procedure bufdump(buf:twbufor;chars:integer);
function fieldtostr(t:byte):string;

//database (manage databases)
procedure find_db_ids(var pp:twdbnr);
procedure destroy_db_ids(var pp:twdbnr);
procedure find_db(db:word;var pp:twdbtype;var pageout:word);
procedure destroy_db_type(var pp:twdbtype);
procedure read_db(page:word;var pp:twdbrecord);
procedure destroy_db_records(var pp:twdbrecord);

//datetime (set/get rex time)
procedure get_rex_time(var y,m,d,h,min,s:word);
procedure set_rex_time(y,m,d,h,min,s:word);


implementation

uses linux;

type
  tpage=record
          len:integer;
	  index:integer;
	  buf:twbufor;
	end;

const
  debug:boolean=false;

var
  tios,oldtios:TermIOS;
  fd:longint;

//open serial port
function open_port(nazwa:string):boolean;
begin
  open_port:=false;
  fd:=fdopen(nazwa,Open_RdWr); // or Open_NoCtty); // or Open_NonBlock);
  if linuxerror<>0 then
  begin
    writeln(stderr);
    writeln(stderr,'Unable to open serial port: ',nazwa);
    writeln(stderr,'Check permissions!');
    exit;
  end;

  if not isatty(fd) then
  begin
    writeln(stderr);
    writeln(stderr,'This is NOT a tty: ',nazwa);
    exit;
  end;
  
  //inicjacja portu
  tcgetattr(fd,oldtios);
  fillbyte(tios,sizeof(tios),0);
  tios.c_cflag:=CS8 or CLOCAL or CREAD or HUPCL;
  tios.c_iflag:= IGNPAR or IGNBRK;
  tios.c_oflag:= 0;
  tios.c_lflag:= 0;
  tios.c_cc[VTIME]:=0;
  tios.c_cc[VMIN]:= 1;
  cfsetispeed(tios,B57600);
  cfsetospeed(tios,B57600);

  tcflush(fd,TCIFLUSH);
  tcsetattr(fd,TCSANOW,tios);
  open_port:=true;
end;

//close serial port
procedure close_port;
begin
  //set speed to 0 and drop DTR
  cfsetospeed(tios, B0);
  cfsetispeed(tios, B0);
  tcflush(fd,TCIFLUSH);
  tcsetattr(fd,TCSANOW,tios);
  //restore old serial line parameters
  tcflush(fd,TCIFLUSH);
  tcsetattr(fd,TCSANOW,oldtios);  
  fdclose(fd);
end;

//add checksum to command/line
function make_cksum(k:string):string;
var
  wynik:string;
  i:byte;
  suma:longint;
begin
  wynik:='';
  suma:=0;
  for i:=1 to length(k) do
  begin
    suma:=suma+ord(k[i]);
    wynik:=wynik+k[i];
  end;
  suma:=suma mod 256;
  make_cksum:=wynik+hexstr(suma,2);
end;

function make_cksum_bin(buf:twbufor;chars:integer):string;
var
  i:integer;
  suma:longint;
begin
  suma:=0;
  for i:=0 to chars-1 do
    suma:=(suma+buf^[i]) mod 256;
  make_cksum_bin:=hexstr(suma,2);
end;

//send command to rex
procedure wyslij_komende(s:string);
begin
  s:=make_cksum(s);
  if debug then writeln(stderr,s);
  s:=#255+#255+s+#10;
  //fcntl(fd,F_SETFL,0);
  fdwrite(fd,s[1],length(s));
  //tcdrain(fd);
  //fcntl(fd,F_SETFL,Open_nonblock);
end;

procedure wyslij_do_portu(buf:pointer;chars:word);
begin
  //fcntl(fd,F_SETFL,0);
  fdwrite(fd,buf,chars);
  //tcdrain(fd);
  //fcntl(fd,F_SETFL,Open_nonblock);
end;

//helper:returns actual time in 1/100s
function podaj_czas:longint;
var
  h,m,s,s100:word;
begin
  gettime(h,m,s,s100);
  podaj_czas:=((h*60+m)*60+s)*100+s100;
end;

//read data from rex
//buf: pointer to array of bytes
//max_chars: max chars to read
//timeout: max time between receiving chars
//returns number of chars readed
function czytaj_wynik(buf:twbufor;max_chars:word;timeout:word):word;
var
  delta,poczatek:longint;
  chars:word;
  b:byte;
  e:longint;
  flags:longint;
begin
  flags:=0;  //it is just for compiler...
  delta:=0;
  chars:=0;
  poczatek:=podaj_czas;
  while (delta<timeout) and (chars<max_chars) do
  begin
    fcntl(fd,F_GETFL,flags);
    fcntl(fd,F_SETFL,flags or Open_nonblock);
    e:=fdread(fd,b,1);
    fcntl(fd,F_SETFL,flags);
    if e>0 then
    begin
      buf^[chars]:=b;
      inc(chars);
      poczatek:=podaj_czas;
    end;
    delta:=podaj_czas-poczatek;
  end;
  czytaj_wynik:=chars;
end;

//helper:converts max 255 first chars from buffer to string
function buftostr(buf:twbufor;chars:word):string;
var
  wynik:string;
begin
  if chars>255 then chars:=255;
  move(buf^[0],wynik[1],chars);
  wynik[0]:=chr(chars);
  buftostr:=wynik;
end;

//helper: dumping buffer
procedure bufdump(buf:twbufor;chars:integer);
var
  wynik,znaki:string;
  linia:byte;
  i:integer;
begin
  linia:=0;
  wynik:='';
  znaki:='';
  for i:=0 to chars-1 do
  begin
    wynik:=wynik+hexstr(buf^[i],2)+' ';
    if (buf^[i]>31) and (buf^[i]<=127) then znaki:=znaki+chr(buf^[i])
      else znaki:=znaki+'.';
    inc(linia);
    if linia>15 then
    begin
      writeln(hexstr(i+1-linia,4),': ',wynik,'  ',znaki);
      linia:=0;
      wynik:='';
      znaki:='';
    end;
  end;
  if linia<>0 then writeln(hexstr(i+1-linia,4),': ',wynik,' ':(16-linia)*3,'  ',znaki);
end;

//read logical page
function dssr(buf:twbufor;page:word):integer;
var
  w,ww:word;
  s:string[5];
  k:integer;
  powtorzenia:byte;
  dobrze:boolean;
begin
  powtorzenia:=0;
  dobrze:=false;
  repeat
    wyslij_komende('DSSR'+hexstr(page,4)+'000002000001');
    repeat
      w:=czytaj_wynik(buf,1,50);
    until (w>0) and ((buf^[0]=ord('O')) or (buf^[0]=ord('N')));
    if buf^[0]=ord('O') then
    begin
      w:=czytaj_wynik(buf,8,50);
      s:=buftostr(buf,5);
      s[1]:='$';
      val(s,w,k);
      //writeln('bajtow::',w);
      ww:=czytaj_wynik(buf,w+3,100);
      dssr:=ww-3;
      if debug then bufdump(buf,w);
      //sprawdzenie
      if (ww-3=w) and (make_cksum_bin(buf,ww-3)=chr(buf^[ww-3])+chr(buf^[ww-2])) then dobrze:=true;
    end else
    begin
      w:=czytaj_wynik(buf,8,50);
      dssr:=-1;
      dobrze:=true;
    end;
    inc(powtorzenia);
  until dobrze or (powtorzenia>9);
  if not dobrze then writeln(stderr,'nie moglem odczytac strony ',page,' z powodu bledow transmisji!');
end;

//read all logical page (linked too!)
function read_page(buf:twbufor;page:word):longint;
var
  tt:twbufor;
  len,l:longint;
  next,prev:word;
begin
  len:=dssr(buf,page);
  next:=buf^[2]+256*buf^[3];
  prev:=buf^[4]+256*buf^[5];
  if len<0 then
  begin
    //error while reading page
    read_page:=-1;
    exit;
  end;
  if prev>0 then
  begin
    //this is not first page
    read_page:=-2;
    exit;
  end;
  if next=0 then
  begin
    //this is single page
    read_page:=len;
    exit;
  end;
  //and there is a multipage...
  getmem(tt,550);
  while next<>0 do
  begin
    l:=dssr(tt,next);
    if l<0 then
    begin
      //error!!!
      freemem(tt,550);
      read_page:=-3;
      exit;
    end;
    move(tt^[6],buf^[len],l-6);
    len:=len+l-6;
    next:=tt^[2]+256*tt^[3];
  end;
  freemem(tt,550);
  read_page:=len;
end;

//write logical page
procedure dssw(buf:twbufor;page:word;chars:integer);
var
  w:word;
  s:string[3];
begin
  if chars<1 then exit;
  wyslij_komende('DSSW'+hexstr(page,4)+'0000'+hexstr(chars,4)+'0001');
  repeat
    w:=czytaj_wynik(buf,5,50);
  until (w>0);
  if debug then bufdump(buf,w);
  s:=make_cksum_bin(buf,chars)+#10;
  //fcntl(fd,F_SETFL,0);
  //for k:=0 to chars-1 do
  fdwrite(fd,buf^[0],chars);
  fdwrite(fd,s[1],3);
  //fcntl(fd,F_SETFL,Open_nonblock);
  repeat
    w:=czytaj_wynik(buf,5,50);
  until (w>0);
  if debug then bufdump(buf,w);
end;

//defrag memory
procedure dssg(buf:twbufor);
var
  bajty:word;
begin
  wyslij_komende('DSSG');
  repeat
    czytaj_wynik(buf,1,500);
    write('.');
  until buf^[0]=ord('O');
  bajty:=czytaj_wynik(buf,4,50);
  if debug then
  begin
    bufdump(buf,bajty);
    writeln(stderr,'razem bajtow:',bajty);
  end;
end;

//format sectors
procedure dssf(buf:twbufor);
var
  bajty:word;
begin
  wyslij_komende('DSSF');
  repeat
    czytaj_wynik(buf,1,5000);
    write('.');
  until buf^[0]=ord('O');
  bajty:=czytaj_wynik(buf,4,50);
  if debug then
  begin
    bufdump(buf,bajty);
    writeln(stderr,'razem bajtow:',bajty);
  end;
end;

//field to string
function fieldtostr(t:byte):string;
var
  temp:string;
begin
  case t of
       1: temp:='DB_INT8';
       2: temp:='DB_INT16';
       3: temp:='DB_INT32';
      11: temp:='DB_FXDSTR';
      12: temp:='DB_FXDDAT';
      21: temp:='DB_VARSTR';
      22: temp:='DB_VARDAT';
      90: temp:='DB_RECID';
  else temp:='DB_UNKNOWN';
  end;
  if debug then temp:=temp+'('+hexstr(t,2)+')';
  fieldtostr:=temp;
end;

//read database ids
procedure find_db_ids(var pp:twdbnr);

procedure read_db_ids_local(pg:word);
var
  debug:boolean;
  recs:word;
  reclen:word;
  page:tpage;
  i:word;
  bytes:integer;
  next,dbid:word;
  bufor:tbufor;
  qq:twdbnr;
begin
  debug:=false;
  //read page and copy to local memory
  bytes:=dssr(@bufor,pg);
  if bytes<0 then
  begin
    writeln(stderr,'trafilem na pusta strone!!!');
    exit;
  end;
  page.len:=bytes;
  page.index:=8;
  getmem(page.buf,bytes);
  move(bufor[0],page.buf^[0],bytes);
  
  //read records
  recs:=page.buf^[6]+256*page.buf^[7];
  if debug then writeln(stderr,'znalazlem ',recs,' na stronie ',pg);
  for i:=1 to recs do
  begin
    reclen:=page.buf^[page.index]+256*page.buf^[page.index+1];
    if debug then writeln(stderr,'dlugosc ',i,' rekordu=',reclen);
    next:=page.buf^[page.index+2]+256*page.buf^[page.index+3];
    if debug then writeln(stderr,'next page:',hexstr(next,4));
    if reclen<=8 then dbid:=0 else
      dbid:=page.buf^[page.index+6]+256*page.buf^[page.index+7];
    if debug then writeln(stderr,'db_id=',dbid,':',hexstr(dbid,4));
    
    if next=0 then
    begin
      new(qq);
      qq^.next:=pp;
      qq^.nr:=dbid;
      pp:=qq;
    end else
    begin
      read_db_ids_local(next);
    end;
    
    page.index:=page.index+reclen;
  end;
  
  freemem(page.buf,bytes);
end;
begin
  pp:=nil;
  read_db_ids_local(1);
end;

//destroy ids structure
procedure destroy_db_ids(var pp:twdbnr);
var
  qq:twdbnr;
begin
  while pp<>nil do
  begin
    if debug then writeln(pp^.nr);
    qq:=pp^.next;
    dispose(pp);
    pp:=qq;
  end;
end;

procedure find_db(db:word;var pp:twdbtype;var pageout:word);
var
  bufor:tbufor;
procedure fdt(pg:word);
var
  debug:boolean;
  recs:word;
  reclen:word;
  page:tpage;
  i,j:word;
  bytes:integer;
  next,dbid:word;
  t:byte;
  tlen:byte;
  fields:word;
  datapage:word;
  tt,ttt:twdbtype;
begin
  debug:=false;
  //read page and copy to local memory
  bytes:=dssr(@bufor,pg);
  if debug then bufdump(@bufor,bytes);
  if bytes<0 then writeln(stderr,'nie moglem odczytac strony: ',pg);
  page.len:=bytes;
  page.index:=8;
  getmem(page.buf,bytes);
  move(bufor[0],page.buf^[0],bytes);
  
  //read records
  recs:=page.buf^[6]+256*page.buf^[7];
  if debug then writeln(stderr,'znalazlem ',recs,' na stronie ',pg);
  for i:=1 to recs do
  begin
    reclen:=page.buf^[page.index]+256*page.buf^[page.index+1];
    if debug then writeln(stderr,'dlugosc ',i,' rekordu=',reclen);
    next:=page.buf^[page.index+2]+256*page.buf^[page.index+3];
    if debug then writeln(stderr,'next page:',hexstr(next,4));
    if reclen<=8 then dbid:=0 else
      dbid:=page.buf^[page.index+6]+256*page.buf^[page.index+7];
    if debug then writeln(stderr,'db_id=',dbid,':',hexstr(dbid,4));
    
    if next=0 then
    begin
      fields:=page.buf^[page.index+22]+256*page.buf^[page.index+23];
      if debug then writeln(stderr,'Baza ',dbid,' ma ',fields,' pol');
      if dbid=db then
      begin
        for j:=1 to fields do
        begin
          t:=page.buf^[page.index+24+j*4];
	  tlen:=page.buf^[page.index+25+j*4];
	  if debug then write(stderr,fieldtostr(t),':',tlen,' ');
	  new(tt);
	  tt^.next:=nil;
	  tt^.typ:=t;
	  tt^.size:=tlen;
	  //dopisanie rekordu na koncu listy
	  if pp=nil then pp:=tt else
	  begin
	    ttt:=pp;
	    while ttt^.next<>nil do
	      ttt:=ttt^.next;
	    ttt^.next:=tt;
	  end;
        end;
        if debug then writeln(stderr);
	datapage:=page.buf^[page.index+28+fields*4]+256*page.buf^[page.index+29+fields*4];
	if debug then writeln(stderr,'Strona z danymi:',hexstr(datapage,2));
        pageout:=datapage;
      end;
    end else
    begin
      fdt(next);
    end;
    
    page.index:=page.index+reclen;
  end;
  
  freemem(page.buf,bytes);
end;
begin
  pp:=nil;
  pageout:=0;
  fdt(1);
end;

procedure destroy_db_type(var pp:twdbtype);
var
  tt:twdbtype;
begin
  while pp<>nil do
  begin
    tt:=pp;
    pp:=pp^.next;
    dispose(tt);
  end;
end;

procedure read_db(page:word;var pp:twdbrecord);
var
  bufor:tbufor;
procedure rdr(pg:word);
var
  debug:boolean;
  recs:word;
  reclen:word;
  page:tpage;
  i:word;
  bytes:integer;
  next:word;
  tt:twdbrecord;
begin
  debug:=false;
  //read page and copy to local memory
  bytes:=read_page(@bufor,pg);
  page.len:=bytes;
  page.index:=8;
  getmem(page.buf,bytes);
  move(bufor[0],page.buf^[0],bytes);
  
  //read records
  recs:=page.buf^[6]+256*page.buf^[7];
  if debug then writeln(stderr,'znalazlem ',recs,' na stronie ',pg);
  for i:=1 to recs do
  begin
    reclen:=page.buf^[page.index]+256*page.buf^[page.index+1];
    if reclen<8 then writeln(stderr,'dlugosc rekordu<8!!!!!');
    if debug then writeln(stderr,'dlugosc ',i,' rekordu=',reclen);
    next:=page.buf^[page.index+2]+256*page.buf^[page.index+3];
    if debug then writeln(stderr,'next page:',hexstr(next,4));
    
    if next=0 then
    begin
      new(tt);
      tt^.next:=pp;
      pp:=tt;
      
      pp^.size:=reclen-8;
      getmem(pp^.data,pp^.size);
      move(page.buf^[page.index+6],pp^.data^[0],pp^.size);
    end else
    begin
      rdr(next);
    end;
    
    page.index:=page.index+reclen;
  end;
  
  freemem(page.buf,bytes);
end;
begin
  pp:=nil;
  rdr(page);
end;

procedure destroy_db_records(var pp:twdbrecord);
var
  tt:twdbrecord;
begin
  while pp<>nil do
  begin
    tt:=pp;
    freemem(pp^.data,pp^.size);
    pp:=pp^.next;
    dispose(tt);
  end;
end;

function open_rex(nazwa:string):boolean;
var
  s:string;
  bufor:tbufor;
  bajty:word;
begin
  open_rex:=false;

  //initialization
  write(stderr,'Opening serial port... ');
  if open_port(nazwa) then writeln(stderr,'done.') else exit;
  s:=#255+#255+#255+#255;
  wyslij_do_portu(@s[1],4);
  close_port;

  //begin of session (BS command)
  open_port(nazwa);
  write(stderr,'Connecting with REX... ');
  wyslij_komende('DSBS0000');
  bajty:=czytaj_wynik(@bufor,10,100);
  if debug then writeln(stderr,buftostr(@bufor,bajty));
  if bajty<>5 then
  begin
    writeln(stderr,'failed!');
    close_port;
    bufdump(@bufor,bajty);
    exit;
  end;
  writeln(stderr,'done.');
  open_rex:=true;
end;

procedure close_rex;
var
  bajty:word;
  bufor:tbufor;
begin
  write(stderr,'Closing connection with REX... ');
  //end of session (BE command);
  wyslij_komende('DSBE');
  bajty:=czytaj_wynik(@bufor,20,50);
  if debug then writeln(stderr,buftostr(@bufor,bajty));
  //closing serial port
  close_port;
  writeln(stderr,'done.');
end;

//time read
function dstr(buf:twbufor):string;
var
  bajty:word;
begin
  write(stderr,'Reading date from REX... ');
  wyslij_komende('DSTR');
  bajty:=czytaj_wynik(buf,29,50);
  //dstr:=buftostr(buf,bajty);
  if (bajty=29) and (make_cksum_bin(buf,26)=chr(buf^[26])+chr(buf^[27]))then
  begin
    dstr:=copy(buftostr(buf,29),3,24);
    writeln(stderr,'done.');
  end else
  begin
    dstr:='';
    writeln(stderr,'failed!');
  end;
end;

procedure dstw(rextime:string);
var
  bajty:word;
  buf:twbufor;
begin
  write(stderr,'Writing date to REX... ');
  getmem(buf,30);
  wyslij_komende('DSTW'+rextime);
  bajty:=czytaj_wynik(buf,29,50);
  if debug then bufdump(buf,bajty);
  if (bajty=5) and (buftostr(buf,4)='OK9A') then writeln(stderr,'done.') else writeln(stderr,'failed!'); 
  freemem(buf,30);
end;

procedure get_rex_time(var y,m,d,h,min,s:word);
var
  buf:twbufor;
  st:string;
  k:integer;
begin
  getmem(buf,30);
  st:=dstr(buf);
  if st<>'' then
  begin
    val('$'+copy(st,1,4),y,k);
    val('$'+copy(st,5,4),m,k);
    val('$'+copy(st,9,4),d,k);
    val('$'+copy(st,13,4),h,k);
    val('$'+copy(st,17,4),min,k);
    val('$'+copy(st,21,4),s,k);
  end else getdatetime(y,m,d,h,min,s);
  freemem(buf,30);
end;

procedure set_rex_time(y,m,d,h,min,s:word);
var
  st:string;
begin
  st:=hexstr(y,4)+hexstr(m,4)+hexstr(d,4)+hexstr(h,4)+hexstr(min,4)+hexstr(s,4);
  dstw(st);
end;


begin
end.
